/* ym68api.c
 *
 * Copyright (C) 2006, bja <bja@kth.se>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ym68api.h"
#include "api68/api68.h"
#include "StSoundLibrary.h"

#define ERROR_STACK_SIZE 8

static int num_errors = 0;
static const char *error_stack[ERROR_STACK_SIZE];

static int get_file_type(const char *filename);
static int open_ym(ym68_t *ym68, const char *filename);
static int open_sc68(ym68_t *ym68, const char *filename);
static int push_error(const char *message);
static const char *pop_error();

ym68_t *ym68_init()
{
    ym68_t *ym68 = (ym68_t *) calloc(1, sizeof (ym68_t));
    if (!ym68) {
	push_error("ym68_init: malloc failed");
	return NULL;
    }
    return ym68;
}

int ym68_quit(ym68_t *ym68)
{
    if (!ym68) {
	return 0;
    }
 
    switch (ym68->type) {

    case YM68_TYPE_YM:
	ymMusicStop((YMMUSIC *) ym68->data);
	ymMusicDestroy((YMMUSIC *) ym68->data);
	break; // FIXME check for errors!

    case YM68_TYPE_SC68:
    case YM68_TYPE_SNDH:
	api68_close((api68_t *) ym68->data);
	api68_shutdown((api68_t *) ym68->data);
	break;

    default:
	push_error("ym68_quit: unknown file type");
	return -1;
    }

    return 0;
}

int ym68_open(ym68_t *ym68, const char *filename)
{
    if (!ym68) {
	push_error("ym68_open: invalid object");
	return -1;
    }

    if (ym68->type != 0) {
	// file is already opened.
	push_error("ym68_open: file is already open");
	return -1;
    }

    switch (get_file_type(filename)) {

    case YM68_TYPE_YM:
	return open_ym(ym68, filename);

    case YM68_TYPE_SC68:
    case YM68_TYPE_SNDH:
	return open_sc68(ym68, filename);

    default:
	ym68->type = YM68_TYPE_UNKNOWN;
	break;
    }

    push_error("ym68_open: unknown file type");
    return -1;
}

int ym68_close(ym68_t *ym68)
{
    if (!ym68) {
	push_error("ym68_close: invalid object");
	return -1;
    }
 
    switch (ym68->type) {

    case YM68_TYPE_YM:
	ymMusicStop((YMMUSIC *) ym68->data);
	ymMusicDestroy((YMMUSIC *) ym68->data);
	ym68->type = 0;
	break; // FIXME check for errors!

    case YM68_TYPE_SC68:
    case YM68_TYPE_SNDH:
	api68_close((api68_t *) ym68->data);
	api68_shutdown((api68_t *) ym68->data);
	ym68->data = NULL;
	ym68->type = 0;
	break;

    default:
	push_error("ym68_close: unknown file type");
	return -1;
    }

    return 0;
}

int ym68_play(ym68_t *ym68, int track) 
{
    if (!ym68) {
	push_error("ym68_play: invalid object");
	return -1;
    }
 
    switch (ym68->type) {

    case YM68_TYPE_YM:
	ymMusicPlay((YMMUSIC *) ym68->data);
	ym68->track = 1;
	break;

    case YM68_TYPE_SC68:
    case YM68_TYPE_SNDH:
	if (api68_play((api68_t *) ym68->data, track)) {
	    push_error("ym68_play: api68_play failed");
	    return -1;
	}
	ym68->track = track;
	break;

    default:
	push_error("ym68_play: unknown file type");
	return -1;
    }

    ym68->state = YM68_STATE_PLAYING;

    return 0;
}

int ym68_process(ym68_t *ym68, void *buf, int n)
{
    int code;

    if (!ym68) {
	push_error("ym68_process: invalid object");
	return -1;
    }
 
    if (ym68->state == YM68_STATE_STOPPED) {
//	push_error("ym68_process: stopped");
	return -1;
    }

    switch (ym68->type) {

    case YM68_TYPE_YM:
	if (ymMusicCompute((YMMUSIC *) ym68->data, (short *) buf, n)) {
	    return 0;
	}
	ym68_stop(ym68);
	return 0;

    case YM68_TYPE_SC68:
    case YM68_TYPE_SNDH:
	code = api68_process((api68_t *) ym68->data, buf, n);
	if (code == API68_MIX_ERROR) {
	    push_error("ym68_process: api68_process failed");
	    printf("ym68_process: %s\n", api68_error());
	    ym68_stop(ym68);
	    return -1;
	}
	if ((code & (API68_LOOP | API68_END)) && ym68->loop) {
	    ym68_stop(ym68);
	    ym68_play(ym68, ym68->track);
	}
	else if ((code & (API68_CHANGE))) {
	    api68_music_info_t info;
	    api68_music_info((api68_t *) ym68->data, &info, -1, 0);
	    ym68->track = info.track;
	}
	else if ((code & API68_END)) {
	    ym68_stop(ym68);
	}
	return code;

    default:
	push_error("ym68_process: unknown file type");
	return -1;
    }

    return 0;
}

int ym68_pause(ym68_t *ym68)
{
    push_error("ym68_pause: not implemented");
    return -1;
}

int ym68_stop(ym68_t *ym68)
{
    if (!ym68) {
	push_error("ym68_stop: invalid object");
	return -1;
    }
 
    switch (ym68->type) {

    case YM68_TYPE_YM:
	ymMusicStop((YMMUSIC *) ym68->data);
	break;

    case YM68_TYPE_SC68:
    case YM68_TYPE_SNDH:
	if (api68_stop((api68_t *) ym68->data)) {
	    push_error("ym68_stop: api68_stop failed");
	    return -1;
	}
	break;

    default:
	push_error("ym68_stop: unknown file type");
	return -1;
    }

    ym68->state = YM68_STATE_STOPPED;

    return 0;
}

int ym68_set_track(ym68_t *ym68, int track)
{
    if (!ym68) {
	push_error("ym68_set_track: invalid object");
	return -1;
    }
 
    ym68->track = track;

    return 0;
}

int ym68_get_track(ym68_t *ym68)
{
    if (!ym68) {
	push_error("ym68_set_track: invalid object");
	return -1;
    }

    return ym68->track;
}

int ym68_get_tracks(ym68_t *ym68)
{
    int tracks = 0;

    if (!ym68) {
	push_error("ym68_get_tracks: invalid object");
	return -1;
    }
 
    switch (ym68->type) {

    case YM68_TYPE_YM:
	return 1;

    case YM68_TYPE_SC68:
    case YM68_TYPE_SNDH:
	tracks = api68_tracks((api68_t *) ym68->data);
	if (tracks < 0) {
	    push_error("ym68_get_tracks: api68_tracks failed");
	    return -1;
	}
	return tracks;

    default:
	push_error("ym68_get_tracks: unknown file type");
	return -1;
    }

    return -1;
}

int ym68_seek(ym68_t *ym68, int pos)
{
    int tracks;
    int track;

    if (!ym68) {
	push_error("ym68_seek: invalid object");
	return -1;
    }
 
    switch (ym68->type) {

    case YM68_TYPE_YM:
	ymMusicSeek((YMMUSIC *) ym68->data, pos);
	return 0;

    case YM68_TYPE_SC68:
    case YM68_TYPE_SNDH:
	tracks = api68_tracks((api68_t *) ym68->data);
	if (tracks < 0) {
	    push_error("ym68_seek: api68_tracks");
	    return -1;
	}
	if (pos > 0) {
	    track = ym68->track + 1;
	    if (track > tracks) {
		track = tracks;
	    }
	}
	else if (pos < 0) {
	    track = ym68->track - 1;
	    if (track < 1) {
		track = 1;
	    }
	}
	else {
	    track = ym68->track;
	}
	api68_stop((api68_t *) ym68->data);
	ym68->track = track;
	api68_play((api68_t *) ym68->data, track);
	return 0;

    default:
	push_error("ym68_seek: unknown file type");
	return -1;
    }

    return -1;
}

int ym68_tell(ym68_t *ym68)
{
    int pos;

    if (!ym68) {
	push_error("ym68_tell: invalid object");
	return -1;
    }
 
    switch (ym68->type) {

    case YM68_TYPE_YM:
	return ymMusicGetPos((YMMUSIC *) ym68->data);

    case YM68_TYPE_SC68:
    case YM68_TYPE_SNDH:
	if (ym68->state == YM68_STATE_STOPPED) {
	    return 0;
	}
	pos = api68_seek((api68_t *) ym68->data, -1);
	if (pos < 0) {
	    push_error("ym68_tell: api68_seek");
	}
	return pos; // FIXME have to keep track of time ourselves.

    default:
	push_error("ym68_tell: unknown file type");
	return -1;
    }

    return -1;
}

int ym68_get_length(ym68_t *ym68)
{
    api68_music_info_t sc68_info;
    ymMusicInfo_t ym_info;

    if (!ym68) {
	push_error("ym68_get_length: invalid object");
	return -1;
    }
 
    switch (ym68->type) {

    case YM68_TYPE_YM:
	ymMusicGetInfo((YMMUSIC *) ym68->data, &ym_info);
	return ym_info.musicTimeInMs;

    case YM68_TYPE_SC68:
    case YM68_TYPE_SNDH:
	if (api68_music_info((api68_t *) ym68->data, &sc68_info, ym68->track, 0) < 0) {
	    push_error("ym68_get_length: api68_music_info failed");
	    return -1;
	}
	return sc68_info.time_ms; 

    default:
	push_error("ym68_get_length: unknown file type");
	return -1;
    }

    return -1;
}

int ym68_loop(ym68_t *ym68, int loop)
{
    if (loop < 0) {
	return ym68->loop;
    }

    if (ym68->type == YM68_TYPE_YM) {
	ymMusicSetLoopMode((YMMUSIC *) ym68->data, loop);
    }

    ym68->loop = loop;

    return 0;
}


int ym68_info(ym68_t *ym68, ym68_info_t *info)
{
    api68_music_info_t sc68_info;
    ymMusicInfo_t ym_info;

    if (!ym68) {
	push_error("ym68_get_length: invalid object");
	return -1;
    }
 
    switch (ym68->type) {

    case YM68_TYPE_YM:
	memset(&ym_info, 0, sizeof (ymMusicInfo_t));
	ymMusicGetInfo((YMMUSIC *) ym68->data, &ym_info);
	info->title = ym_info.pSongName;
	info->author = ym_info.pSongAuthor;
	info->comment = ym_info.pSongComment;
	info->type = ym_info.pSongType;
	info->player = ym_info.pSongPlayer;
	info->time_ms = ym_info.musicTimeInMs;
	info->tracks = 1;
	info->track = ym68->track;
	info->loop = ym68->loop;
	return 0;

    case YM68_TYPE_SC68:
    case YM68_TYPE_SNDH:
	memset(&sc68_info, 0, sizeof (api68_music_info_t));
	if (api68_music_info((api68_t *) ym68->data, 
			     &sc68_info, ym68->track, 0) < 0) {
	    push_error("ym68_get_info: api68_music_info failed");
	    return -1;
	}
	info->title = sc68_info.title;
	info->author = sc68_info.author;
	info->composer = sc68_info.composer;
	info->player = sc68_info.replay;
	info->type = sc68_info.hwname;
	info->tracks = sc68_info.tracks;
	info->comment = " ";
	info->time_ms = sc68_info.time_ms;
	info->track = sc68_info.track;
	info->loop = ym68->loop;
	return 0; 

    default:
	push_error("ym68_get_length: unknown file type");
	return -1;
    }

    return -1;
}

static int push_error(const char *message)
{
//#ifdef _DEBUG
    printf(message);
    printf("\n");
//#endif
    if (num_errors == ERROR_STACK_SIZE - 1) {
	error_stack[num_errors] = "push_error: error_stack overflow";
	return -1;
    }
    error_stack[num_errors] = message;
    num_errors++;
    return 0;
}

static const char *pop_error()
{
    if (num_errors >= ERROR_STACK_SIZE) {
	return "pop_error: invalid error stack";
    }
    if (num_errors > 0) {
	return error_stack[num_errors--];
    }
    return "";
}


static int get_file_type(const char *filename)
{
    int len = strlen(filename);

    if (len < 3) {
	return YM68_TYPE_UNKNOWN;
    }

    if (!strncasecmp(".ym", &filename[len-3], 3)) {
	return YM68_TYPE_YM;
    }

    if (!strncasecmp(".snd", &filename[len-4], 4)) {
	return YM68_TYPE_SNDH;
    }

    if (len > 4) {

	if (!strncasecmp(".sc68", &filename[len-5], 5)) {
	    return YM68_TYPE_SC68;
	}

	if (!strncasecmp(".sndh", &filename[len-5], 5)) {
	    printf("--foo--\n");
	    return YM68_TYPE_SNDH;
	}
    }

    return YM68_TYPE_UNKNOWN;
}

static int open_ym(ym68_t *ym68, const char *filename)
{
    ymMusicInfo_t yminfo;

    if (!ym68) {
	push_error("open_ym: ym68 is null");
	return -1;
    }
  
    YMMUSIC *ym = ymMusicCreate();

    if (!ymMusicLoad(ym, filename)) {
	push_error("open_ym: ymMusicLoad failed");
	ymMusicDestroy(ym);
	return -1;
    }

    ym68->data = ym;
    ym68->type = YM68_TYPE_YM;
    ym68->track = 1;

    ymMusicSetLoopMode(ym, YMFALSE);

    ymMusicGetInfo(ym, &yminfo);

    printf("Title: %s\nAuthor: %s\nComment: %s\n",
	   yminfo.pSongName,
	   yminfo.pSongAuthor,
	   yminfo.pSongComment);

    printf("Total music time: %d seconds.\n",
	   yminfo.musicTimeInSec);

    return 0;
}

static int open_sc68(ym68_t *ym68, const char *filename)
{
    api68_t *sc68;
    api68_init_t init68;

    printf("filename: %s\n", filename);

    memset(&init68, 0, sizeof (init68));
    init68.alloc = malloc;
    init68.free = free;

    sc68 = api68_init(&init68);
    if (!sc68) {
	push_error("open_sc68: api68_init failed");
	fprintf(stderr,api68_error());
	return -1;
    }

    if (api68_verify_file(filename) < 0) {
	push_error("open_sc68: api68_verify_file failed");
	fprintf(stderr,api68_error());
	return -1;
    }

    if (api68_load_file(sc68, filename)) {
	push_error("open_sc68: api68_load_file failed");
	return -1;
    }


    ym68->type = YM68_TYPE_SC68;//get_file_type(filename);
    ym68->data = sc68;

    return 0;
}
