/* ym68sdl.c
 *
 * Copyright (C) 2006, bja <bja@kth.se>
 *
 * This program is based on Aylet.gp2x by ParkyDR.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <signal.h>
#include <getopt.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>

#include "ym68api.h"
#include "ym68sdl.h"
#include "ym68gui.h"

#define VERSION_STRING "version 0.3b2"

#define MAX(x,y) ((x) < (y) ? (y) : (x))
#define MIN(x,y) ((x) > (y) ? (y) : (x))

// shared with gui

ym68_info_t ym68info;		// used by the gui to show info about file
int ymfile;			// index of current playing file in paylist
int playlist_enable = 1;	// enable / disable playlist mode
int loop_enable = 0;		// enable / disable loop mode
int update_flag = 0;

// global variables

static int quit;		// exit button pressed (quit program)
static Uint8 *audio_buffer;	// buffer used by SDL_MixAudio (for volume)
static ym68_t *ym68 = 0;	// currently loaded ym/sc68/sndh file


static Uint32 track_position = 0;
static SDL_AudioSpec audio_format;

// the following are configurable through command line options

static int audio_buffer_size = 8192; // number of 16-bit stereo samples
static int common_volume = SDL_MIX_MAXVOLUME / 2;
static int adjusted_volume = 0;		// the volume after ym adjustment
static int ym_volume_adjust = 20;	// ym tunes are played this much louder
static int frame_delay = 50;		// time in ms to sleep between frames

// function prototypes

static void init_audio(int channels);
static int load_ym68_file(const char *path);


/* Display informatino about the specified track of the currently
 * loaded file to stdout.
 */
static void print_ym68_info(int track)
{
    ym68_info_t info;
    if (!ym68_info(ym68,&info)) {
	const char *type = ym68->type == YM68_TYPE_SC68 ? info.composer : " ";
	printf("-- Track      : %d/%d\n",info.track, info.tracks);
	printf("-- Title      : %s\n",info.title);
	printf("-- Author     : %s\n",info.author);
	printf("-- Composer   : %s\n",type);
	printf("-- player     : %s\n",info.player);
	printf("-- Hardware   : %s\n",info.type);
	printf("-- Duration   : %d\n", info.time_ms/1000);
    }
}

static void audio_callback(void *unused, Uint8 *stream, int len)
{
    if (len > audio_buffer_size << 2) {
	fprintf(stderr,"audio_callback: buffer overflow");
	len = audio_buffer_size << 2;
    }
    int num_samples = len>>ym68->channels;
    int code = ym68_process(ym68, audio_buffer, num_samples);
    if (code & 0x14) {
	update_flag = 1;
    }
    SDL_MixAudio(stream, audio_buffer, len, adjusted_volume);
}

static void adjust_volume(void)
{
    adjusted_volume = common_volume;
    if (ym68->type != YM68_TYPE_YM) {
	adjusted_volume -= ym_volume_adjust;
    }
    adjusted_volume = MIN(SDL_MIX_MAXVOLUME, adjusted_volume);
    adjusted_volume = MAX(0, adjusted_volume);
}

/* do action requested by ui code. returns zero if we need to
 * exit the current track.
 */
int action_callback(enum cb_action_tag action)
{
    switch(action) {

    case cb_quit:
	quit = 1;
	return 0;	/* exit track */
    
    case cb_highspeed:
	playlist_enable = !playlist_enable;
	break;

    case cb_prev_file:
	ay_file = ay_file - 1;
	if (ay_file < 0) {
	    ay_file += ay_num_files;
	}
	break;

    case cb_next_file:
	ay_file = (ay_file + 1) % ay_num_files;
	break;

    case cb_prev_track:
	if (ym68->type == YM68_TYPE_YM) {
	    ym68_seek(ym68, track_position - 10000);
	}
	else {
	    ym68_seek(ym68, -1);
	}
	break;

    case cb_next_track:
	if (ym68->type == YM68_TYPE_YM) {
	    ym68_seek(ym68, track_position + 10000);
	}
	else {
	    ym68_seek(ym68, 1);
	}
	break;

    case cb_play:
    case cb_stop:
	SDL_PauseAudio(1);
	SDL_LockAudio();
	playing = 0;
	ym68_stop(ym68);
	if (load_ym68_file(ay_filenames[ay_file]) < 0) {
	    break;
	}
	adjust_volume();
	printf("\n-- Playing tune from \"%s\"\n", ay_filenames[ay_file]);
	print_ym68_info(ym68->track);
	if (ym68_play(ym68,ym68->track) < 0) {
	    break;
	}
	playing = 1;
	paused = 0;
	SDL_UnlockAudio();
	SDL_PauseAudio(0);
	break;

    case cb_pause:
	paused = !paused;
	SDL_PauseAudio(paused);
	break;

    case cb_restart:
	loop_enable = !loop_enable;
	ym68_loop(ym68, loop_enable);
	break;
    
    case cb_dec_stopafter:
	break;

    case cb_inc_stopafter:
	break;

    case cb_dec_fadetime:
	common_volume = MAX(common_volume-5, 0);
	adjust_volume();
	break;
    
    case cb_inc_fadetime:
	common_volume = MIN(common_volume+5, SDL_MIX_MAXVOLUME);
	adjust_volume();
	break;

    case cb_dec_vol:
    case cb_inc_vol:
	/* XXX NYI */
	break;

    default:	/* for cb_none */
	break;
    }

    ym68_info(ym68, &ym68info);

    return(1);
}

static void cleanup(void)
{
    SDL_PauseAudio(1);
    SDL_CloseAudio();

    if (ym68) {
	ym68_quit(ym68);
	ym68 = NULL;
    }
    if (audio_buffer) {
	free(audio_buffer);
	audio_buffer = NULL;
    }
}

static int load_ym68_file(const char *path)
{
    ym68_close(ym68);

    if (ym68_open(ym68, path)) {
	printf("load_ym68_file: error ym68_open\n");
	return -1;
    }
    
    ym68_info(ym68, &ym68info);
    ym68_loop(ym68, loop_enable);
    ym68->track = 1;

    if (ym68->type == YM68_TYPE_YM) {
	init_audio(1);
	ym68->channels = 1;

    }
    else {
	init_audio(2);
	ym68->channels = 2;
    }

    ymfile = ay_file;
    
    return 0;
}

/*
 * Initializes the audio device.
 */
static void init_audio(int channels)
{
    SDL_AudioSpec res;
    if (audio_format.channels == channels) {
	return;
    }

    if (audio_format.channels == 0) {
	if (SDL_InitSubSystem(SDL_INIT_AUDIO)) {
	    fprintf(stderr, "Unable to initialize audio: %s\n",
		    SDL_GetError());
	    exit(1);
	}
	audio_buffer = (Uint8 *) calloc(1, audio_buffer_size << 2);
	if (!audio_buffer) {
	    fprintf(stderr, "Unable to initialize audio: calloc failed\n");
	    exit(1);
	}
    }
    else {
	SDL_PauseAudio(1);
	SDL_CloseAudio();
    }

    audio_format.freq = 44100;
    audio_format.format = AUDIO_S16;
    audio_format.channels = channels;
    audio_format.samples = audio_buffer_size;
    audio_format.callback = audio_callback;
    audio_format.userdata = NULL;

    if (SDL_OpenAudio(&audio_format, &res) < 0) {
	fprintf(stderr, "Unable to open audio: %s\n", SDL_GetError());
	exit(1);
    }

    printf("-- audio_init: freq: %d\n", audio_format.freq);
    printf("-- audio_init: format: %d\n", audio_format.format);
    printf("-- audio_init: channels: %d\n", audio_format.channels);
    printf("-- audio_init: samples: %d\n", audio_format.samples);
}

/* This function filters the result from scandir to only show the
 * supported file types in the playlist.
 */
static int filter(const struct dirent *entry)
{
    int len = strlen(entry->d_name);

    if (len < 3) {
	return 0;
    }
    if (!strncasecmp(&entry->d_name[len-3],".ym",3)) {
	return 1;
    }
    if (len < 4) {
	return 0;
    }
    if (!strncasecmp(&entry->d_name[len-4],".snd",4)) {
	return 1;
    }
    if (len < 5) {
	return 0;
    }
    if (!strncasecmp(&entry->d_name[len-5],".sc68",5)) {
	return 1;
    }
    if (!strncasecmp(&entry->d_name[len-5],".sndh",5)) {
	return 1;
    }
    return 0;
}

/* This function adds the supported files found in the specified
 * directory to the playlist.
 */
static void list_ym68_files(const char *path)
{
    struct dirent **entries = NULL;
    char **filenames;
    int nfiles = 0;
    int i;

    nfiles = scandir(path, &entries, filter, alphasort);
    if (nfiles < 0) {
	return;
    }

    filenames = (char **) malloc(nfiles * sizeof (char *));
    if (!filenames) {
	fprintf(stderr, "list_ym_files: Unable to allocate memory!\n");
    }

    for (i=0; i<nfiles; i++) {
	int len = strlen(path) + strlen(entries[i]->d_name) + 2;
	filenames[i] = (char *) malloc(len);
	snprintf(filenames[i], len, "%s/%s", path, entries[i]->d_name);
	//fprintf(stderr, "%s\n", entries[i]->d_name);
    }

    if (ay_filenames != NULL) {
	for (i=0; i<ay_num_files; i++) {
	    free(ay_filenames[i]);
	}
	ay_filenames = NULL;
	ay_num_files = 0;
    }

    ay_filenames = filenames;
    ay_num_files = nfiles;

    // XXX free entries?
}

static int run(void)
{

    while (!quit) {

	if (playing) {

	    ym68_info(ym68, &ym68info);
	    track_position = ym68_tell(ym68);
	    tunetime.min = track_position / 60000;
	    tunetime.sec = (track_position / 1000) % 60;

	    if (ym68->state == YM68_STATE_STOPPED) {
		if (playlist_enable) {
		    SDL_PauseAudio(1);
		    SDL_LockAudio();
		    ay_file = (ay_file + 1) % ay_num_files;
		    load_ym68_file(ay_filenames[ay_file]);
		    ym68_play(ym68, 1);
		    SDL_UnlockAudio();
		    SDL_PauseAudio(0);
		}
		else {
		    SDL_PauseAudio(1);
		    playing = 0;
		}

	    }
	}

	ui_frame();

	SDL_Delay(frame_delay);
    }

    return 0;
}

static void usage(char *progname)
{
    char *name = strdupa(progname);
    printf("Usage: %s [OPTIONS] PATH\n\n", basename(name));
    printf("Where options include:\n");
    printf("  -b       sample buffer size [1024-16384]\n");
    printf("  -v       initial playback volume [0-128]\n");
    printf("  -y       adjustment of the ym playback volume [0-128]\n");
    printf("  -f       frame delay in ms [0-500]\n");
    printf("  -h       display help screen\n\n");
    printf("PATH specifies where to look for music files.\n\n");
    printf("Report bugs to <bja@kth.se>.\n");
    exit(1);
}

int main(int argc,char *argv[])
{
    /* parse command line arguments */

    int optc;

    for (;;) {

	optc = getopt(argc, argv, "b:y:f:v:h");

	if (optc == -1) {
	    break;
	}

	switch (optc) {

	case 'b':
	    audio_buffer_size = atoi(optarg);
	    break;

	case 'y':
	    ym_volume_adjust = atoi(optarg);
	    break;

	case 'f':
	    frame_delay = atoi(optarg);
	    break;    

	case 'v':
	    common_volume = atoi(optarg);
	    break;    

	case 'h':
	    usage(argv[0]);
	    exit(1);

	default:
	    usage(argv[0]);
	}
    }
    
    if ((optind == argc) || (argc - optind != 1)) {
	usage(argv[0]);
	return 1;
    }

    if (atexit(cleanup)) {
	fprintf(stderr, "Unable to install atexit handler!\n");
	exit(1);
    }

    printf("%s %s\n", basename(argv[0]), VERSION_STRING);
    printf("Copyright (C) 2006, bja.\n");
    printf("This is free software.  You may redistribute copies "
	   "of it under the terms of\nthe GNU General Public License "
	   "<http://www.gnu.org/licenses/gpl.html>.\n"
	   "There is NO WARRANTY, to the extent permitted by law.\n\n");

    printf("-- Buffer size: %d\n", audio_buffer_size);
    printf("-- Frame delay: %d\n", frame_delay);
    printf("-- Volume:      %d\n", common_volume);
    printf("-- YM volume adjustment: %d\n", ym_volume_adjust);

    ym68 = ym68_init();
    list_ym68_files(argv[optind]);
    load_ym68_file(ay_filenames[ay_file]);

    playing = 0;

    ui_init(argc, argv);

    run();

    ui_end(); // does nothing

    return 0;
}
